-- SPDX-FileCopyrightText: 2021 Oxhead Alpha
-- SPDX-License-Identifier: LicenseRef-MIT-OA

{- | Module, containing restrictions imposed by instruction or value scope.

Michelson have multiple restrictions on values, examples:

* @operation@ type cannot appear in parameter.
* @big_map@ type cannot appear in @PUSH@-able constants.
* @contract@ type cannot appear in type we @UNPACK@ to.

Thus we declare multiple "scopes" - constraints applied in corresponding
situations, for instance

* 'ParameterScope';
* 'StorageScope';
* 'ConstantScope'.

Also we separate multiple "classes" of scope-related constraints.

* 'ParameterScope' and similar ones are used within Michelson engine,
they are understandable by GHC and produce human-readable errors.

* Lorentz and other eDSLs may declare their own constraints, in most cases
you should use them. For example see @Lorentz.Constraints@ module.

-}

module Morley.Michelson.Typed.Scope
  ( -- * Scopes
    ConstantScope
  , DupableScope
  , StorageScope
  , PackedValScope
  , ParameterScope
  , UntypedValScope
  , UnpackedValScope
  , ViewableScope
  , ComparabilityScope

  , IsDupableScope

  , BadTypeForScope (..)
  , CheckScope (..)
  , WithDeMorganScope (..)
  , deMorganForbidT
  , Comparable
  , comparableImplies
  , WellTyped
  , NotWellTyped (..)

    -- * Implementation internals
  , ContainsT
  , IsComparable

  , ForbidT
  , ForbidOp
  , ForbidContract
  , ForbidTicket
  , ForbidBigMap
  , ForbidNestedBigMaps
  , ForbidNonComparable

  , TPresence(..
      , OpPresent, OpAbsent
      , ContractPresent, ContractAbsent
      , TicketPresent, TicketAbsent
      , BigMapPresent, BigMapAbsent
      , NestedBigMapsPresent, NestedBigMapsAbsent
      , SaplingStatePresent, SaplingStateAbsent
      )
  , TPredicateSym(..)
  , SingTPredicateSym(..)
  , checkTPresence
  , tPresence
  , tAbsence

  , Comparability(..)
  , checkComparability
  , comparabilityPresence

  , getWTP
  , getWTP'
    -- * Re-exports
  , withDict
  , SingI (..)
  , (:-)(..)
    -- * Deprecated
  , OpPresence
  , ContractPresence
  , TicketPresence
  , BigMapPresence
  , NestedBigMapsPresence
  , ProperParameterBetterErrors
  , ProperStorageBetterErrors
  , ProperConstantBetterErrors
  , ProperDupableBetterErrors
  , ProperPackedValBetterErrors
  , ProperUnpackedValBetterErrors
  , ProperUntypedValBetterErrors
  , ProperViewableBetterErrors
  , ProperNonComparableValBetterErrors
  , opAbsense
  , contractTypeAbsense
  , bigMapAbsense
  , nestedBigMapsAbsense
  , checkOpPresence
  , checkContractTypePresence
  , checkTicketPresence
  , checkBigMapPresence
  , checkNestedBigMapsPresence
  , HasNoBigMap
  , HasNoNestedBigMaps
  , HasNoOp
  , HasNoContract
  , HasNoTicket
  , ContainsBigMap
  , ContainsContract
  , ContainsNestedBigMaps
  , ContainsOp
  , ContainsTicket
  , comparabilityImpliesNoNestedBigMaps
  , comparabilityImpliesNoOp
  , getComparableProofS
  ) where

import Data.Constraint (withDict, (:-)(..))
import Data.Singletons (SingI(..))

import Morley.Michelson.Typed.Scope.Internal.CheckScope
import Morley.Michelson.Typed.Scope.Internal.Comparable
import Morley.Michelson.Typed.Scope.Internal.Deprecated
import Morley.Michelson.Typed.Scope.Internal.ForbidT
import Morley.Michelson.Typed.Scope.Internal.Presence
import Morley.Michelson.Typed.Scope.Internal.Scopes
import Morley.Michelson.Typed.Scope.Internal.WellTyped
import Morley.Michelson.Typed.Scope.Internal.WithDeMorganScope
